// SPDX-License-Identifier: AGPL-3.0-or-later
use crate::{msg::*, server::FsProvider};

use btlib::{
    accessor::Accessor,
    bterr,
    crypto::{Creds, Decrypter, Signer},
    error::{BtErr, DisplayErr},
    AuthzAttrs, BlockAccessor, BlockError, BlockMeta, BlockMetaSecrets, BlockOpenOptions,
    BlockPath, BlockReader, DirEntry, Directory, Epoch, FileBlock, FlushMeta, IssuedProcRec,
    MetaAccess, MetaReader, Positioned, Principal, Principaled, ProcRec, Result, Split, TrySeek,
    WriteDual, ZeroExtendable,
};
use btserde::{read_from, write_to};
use core::future::{ready, Ready};
use log::{debug, error, warn};
use positioned_io::Size;
use serde::{Deserialize, Serialize};
use std::{
    collections::hash_map::{self, HashMap},
    fmt::{Display, Formatter},
    fs::File,
    io::{self, Read as IoRead, Seek, SeekFrom, Write as IoWrite},
    net::{IpAddr, Ipv6Addr},
    path::{Path, PathBuf},
    sync::{
        atomic::{AtomicU64, Ordering},
        Arc, Mutex, RwLock, RwLockWriteGuard,
    },
    time::Duration,
};

pub use private::{Authorizer, AuthzContext, Error, LocalFs, ModeAuthorizer};

mod private {
    use super::*;

    type Inode = u64;
    type Handle = u64;

    #[derive(Debug, Clone, PartialEq, Eq, Hash, Serialize, Deserialize)]
    pub enum Error {
        NotOpen(Inode),
        InvalidHandle { inode: Inode, handle: Handle },
        NoHandlesAvailable(Inode),
        InodeNotFound(Inode),
        ReadOnlyHandle(Handle),
        WrongOwner,
    }

    impl Display for Error {
        fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
            match self {
                Error::NotOpen(inode) => write!(f, "inode {inode} is not open"),
                Error::InvalidHandle { inode, handle } => {
                    write!(f, "invalid handle {handle} for inode {inode}")
                }
                Error::NoHandlesAvailable(inode) => {
                    write!(f, "no handles are available for inode {inode}")
                }
                Error::InodeNotFound(inode) => write!(f, "inode {inode} could not be found"),
                Error::ReadOnlyHandle(handle) => {
                    write!(f, "cannot mutably access read-only handle {handle}")
                }
                Error::WrongOwner => write!(f, "handle is not owned by the requestor"),
            }
        }
    }

    trait BlockPathExt {
        fn assert_eq(&self, other: &BlockPath) -> Result<()>;
    }

    impl BlockPathExt for BlockPath {
        fn assert_eq(&self, other: &BlockPath) -> Result<()> {
            if self != other {
                Err(Error::WrongOwner.into())
            } else {
                Ok(())
            }
        }
    }

    impl std::error::Error for Error {}

    /// This type provides context for an authorization decision as to whether a given process will
    /// be allowed to access a block.
    pub struct AuthzContext<'a> {
        /// The path from which this request was received.
        pub from: &'a BlockPath,
        /// The attributes of the principal whose access is being authorized.
        pub attrs: &'a AuthzAttrs,
        /// A reference to the metadata of a block, the access to which is being authorized.
        pub meta: &'a BlockMeta,
    }

    impl<'a> AuthzContext<'a> {
        fn new(from: &'a BlockPath, attrs: &'a AuthzAttrs, meta: &'a BlockMeta) -> Self {
            Self { from, attrs, meta }
        }
    }

    /// A trait for types which can render authorization decisions.
    pub trait Authorizer {
        /// Returns [Ok] if read authorization is granted, and [Err] otherwise.
        fn can_read(&self, ctx: &AuthzContext<'_>) -> io::Result<()>;
        /// Returns [Ok] if write authorization is granted, and [Err] otherwise.
        fn can_write(&self, ctx: &AuthzContext<'_>) -> io::Result<()>;
        /// Returns [Ok] if execute authorization is granted, and [Err] otherwise.
        fn can_exec(&self, ctx: &AuthzContext<'_>) -> io::Result<()>;
    }

    /// A particularly simple authorizer that just looks at the mode bits in the block metadata
    /// to make authorization decisions.
    pub struct ModeAuthorizer;

    impl ModeAuthorizer {
        fn authorize(mode: u32, mask: u32, denied_msg: &str) -> io::Result<()> {
            if (mode & mask) != 0 {
                Ok(())
            } else {
                Err(io::Error::new(io::ErrorKind::PermissionDenied, denied_msg))
            }
        }

        fn user_is_root(ctx: &AuthzContext<'_>) -> bool {
            ctx.attrs.uid == 0
        }
    }

    impl Authorizer for ModeAuthorizer {
        fn can_read(&self, ctx: &AuthzContext<'_>) -> io::Result<()> {
            if Self::user_is_root(ctx) {
                return Ok(());
            }
            let secrets = ctx.meta.body().secrets()?;
            let mask = (libc::S_IRUSR * (secrets.uid == ctx.attrs.uid) as u32)
                | (libc::S_IRGRP * (secrets.gid == ctx.attrs.gid) as u32)
                | libc::S_IROTH;
            Self::authorize(secrets.mode, mask, "read access denied")
        }

        fn can_write(&self, ctx: &AuthzContext<'_>) -> io::Result<()> {
            if Self::user_is_root(ctx) {
                return Ok(());
            }
            let secrets = ctx.meta.body().secrets()?;
            let mask = (libc::S_IWUSR * (secrets.uid == ctx.attrs.uid) as u32)
                | (libc::S_IWGRP * (secrets.gid == ctx.attrs.gid) as u32)
                | libc::S_IWOTH;
            Self::authorize(secrets.mode, mask, "write access denied")
        }

        fn can_exec(&self, ctx: &AuthzContext<'_>) -> io::Result<()> {
            if Self::user_is_root(ctx) {
                return Ok(());
            }
            let secrets = ctx.meta.body().secrets()?;
            let mask = (libc::S_IXUSR * (secrets.uid == ctx.attrs.uid) as u32)
                | (libc::S_IXGRP * (secrets.gid == ctx.attrs.gid) as u32)
                | libc::S_IXOTH;
            Self::authorize(secrets.mode, mask, "exec access denied")
        }
    }

    type HandleValueParts<'a> = (
        &'a Mutex<Option<Accessor<&'static [u8]>>>,
        &'a Arc<BlockPath>,
        Flags,
    );

    enum HandleValue {
        File {
            accessor: Mutex<Option<Accessor<&'static [u8]>>>,
            owner: Arc<BlockPath>,
            flags: Flags,
        },
        Directory {
            accessor: Mutex<Option<Accessor<&'static [u8]>>>,
            owner: Arc<BlockPath>,
            flags: Flags,
            dir: Directory,
        },
    }

    impl HandleValue {
        fn new<T: Size>(accessor: Accessor<T>, owner: Arc<BlockPath>, flags: Flags) -> HandleValue {
            let (accessor, ..) = accessor.split();
            HandleValue::File {
                accessor: Mutex::new(Some(accessor)),
                owner,
                flags,
            }
        }

        fn accessor(&self) -> &Mutex<Option<Accessor<&'static [u8]>>> {
            match self {
                Self::File { accessor, .. } => accessor,
                Self::Directory { accessor, .. } => accessor,
            }
        }

        fn flags(&self) -> Flags {
            match self {
                Self::File { flags, .. } => *flags,
                Self::Directory { flags, .. } => *flags,
            }
        }

        fn parts(&self) -> HandleValueParts<'_> {
            match self {
                Self::File {
                    accessor,
                    owner,
                    flags,
                } => (accessor, owner, *flags),
                Self::Directory {
                    accessor,
                    owner,
                    flags,
                    ..
                } => (accessor, owner, *flags),
            }
        }

        fn use_accessor<T, F: FnOnce(Accessor<&'static [u8]>) -> (Accessor<&'static [u8]>, T)>(
            &self,
            cb: F,
        ) -> Result<T> {
            let mut guard = self.accessor().lock().display_err()?;
            let accessor = guard
                .take()
                .ok_or_else(|| bterr!("accessor has already been taken"))?;
            let (accessor, output) = cb(accessor);
            *guard = Some(accessor);
            Ok(output)
        }

        fn convert_to_dir<C: Signer + Principaled + Decrypter>(
            self,
            block: &mut FileBlock<C>,
        ) -> io::Result<HandleValue> {
            let (accessor, owner, flags) = match self {
                Self::File {
                    accessor,
                    owner,
                    flags,
                } => (accessor, owner, flags),
                Self::Directory {
                    accessor,
                    owner,
                    flags,
                    ..
                } => (accessor, owner, flags),
            };
            let accessor = accessor
                .into_inner()
                .display_err()?
                .ok_or_else(|| bterr!("a thread paniced while it held the accessor lock"))?;
            let mut accessor = Accessor::combine(accessor, block);
            let dir = accessor.read_dir()?;
            let (accessor, ..) = accessor.split();
            Ok(HandleValue::Directory {
                dir,
                accessor: Mutex::new(Some(accessor)),
                owner,
                flags,
            })
        }

        fn directory(&self) -> io::Result<&Directory> {
            match self {
                Self::Directory { dir, .. } => Ok(dir),
                _ => Err(io::Error::new(
                    io::ErrorKind::Other,
                    "handle is not for a directory",
                )),
            }
        }

        fn access_block<B: Size, T, F: FnOnce(&mut Accessor<B>, Flags) -> Result<T>>(
            &self,
            from: &BlockPath,
            block: B,
            cb: F,
        ) -> Result<T> {
            let (_, owner, flags) = self.parts();
            owner.assert_eq(from)?;
            self.use_accessor(|accessor| {
                let mut accessor = Accessor::combine(accessor, block);
                let result = cb(&mut accessor, flags);
                let (accessor, ..) = accessor.split();
                (accessor, result)
            })?
        }

        fn set_flags(&mut self, new_flags: Flags) {
            match self {
                Self::File { flags, .. } => *flags = new_flags,
                Self::Directory { flags, .. } => *flags = new_flags,
            }
        }
    }

    struct InodeTableValue<C> {
        block: Accessor<FileBlock<C>>,
        handle_values: HashMap<Handle, HandleValue>,
        next_handle: Handle,
        lookup_counts: HashMap<Arc<BlockPath>, u64>,
        delete: bool,
    }

    impl<C: Signer + Principaled + Decrypter> InodeTableValue<C> {
        fn new(block: Accessor<FileBlock<C>>, opener: Arc<BlockPath>) -> InodeTableValue<C> {
            let mut lookup_counts = HashMap::with_capacity(1);
            lookup_counts.insert(opener, 1);
            Self {
                block,
                handle_values: HashMap::new(),
                next_handle: 1,
                lookup_counts,
                delete: false,
            }
        }

        fn invalid_handle_err(handle: Handle) -> io::Error {
            io::Error::new(io::ErrorKind::Other, format!("invalid handle {handle}"))
        }

        fn value(&self, handle: Handle) -> io::Result<&HandleValue> {
            self.handle_values
                .get(&handle)
                .ok_or_else(|| Self::invalid_handle_err(handle))
        }

        fn block(&self) -> &FileBlock<C> {
            self.block.get_ref()
        }

        fn block_mut(&mut self) -> &mut FileBlock<C> {
            self.block.get_mut()
        }

        fn convert_to_dir(&mut self, handle: Handle) -> io::Result<()> {
            let value = self
                .handle_values
                .remove(&handle)
                .ok_or_else(|| Self::invalid_handle_err(handle))?;
            let block = self.block_mut();
            let value = value.convert_to_dir(block)?;
            self.handle_values.insert(handle, value);
            Ok(())
        }

        fn access_block<T, F: FnOnce(&mut Accessor<&FileBlock<C>>, Flags) -> Result<T>>(
            &self,
            from: &BlockPath,
            handle: Handle,
            cb: F,
        ) -> Result<T> {
            let value = self.value(handle)?;
            let block = self.block();
            value.access_block(from, block, cb)
        }

        fn access_block_mut<T, F: FnOnce(&mut Accessor<&mut FileBlock<C>>, Flags) -> Result<T>>(
            &mut self,
            from: &BlockPath,
            handle: Handle,
            cb: F,
        ) -> Result<T> {
            let value = self
                .handle_values
                .get(&handle)
                .ok_or_else(|| Self::invalid_handle_err(handle))?;
            if !value.flags().writeable() {
                return Err(Error::ReadOnlyHandle(handle).into());
            }
            let inner = self.block.get_mut();
            value.access_block(from, inner, cb)
        }

        fn borrow_block<T, F: FnOnce(&Accessor<FileBlock<C>>) -> Result<T>>(
            &self,
            cb: F,
        ) -> Result<T> {
            cb(&self.block)
        }

        fn borrow_block_mut<T, F: FnOnce(&mut Accessor<FileBlock<C>>) -> Result<T>>(
            &mut self,
            cb: F,
        ) -> Result<T> {
            cb(&mut self.block)
        }

        fn new_handle(&mut self, owner: Arc<BlockPath>, flags: Flags) -> Result<Handle> {
            if self.handle_values.len() as u64 >= u64::MAX {
                return Err(bterr!("no handles are available"));
            }
            let mut handle_value = HandleValue::new(Accessor::new(self.block())?, owner, flags);
            loop {
                let handle = self.next_handle;
                self.next_handle = self.next_handle.wrapping_add(1);
                match self.handle_values.insert(handle, handle_value) {
                    Some(prev) => {
                        // We've wrapped around and this handle is already taken. Put the previous
                        // value back and try again.
                        handle_value = self.handle_values.insert(handle, prev).unwrap();
                    }
                    // We generated an unused handle. Return it.
                    None => return Ok(handle),
                }
            }
        }

        fn set_flags(&mut self, handle: Handle, flags: Flags) -> Result<()> {
            let handle_value = self
                .handle_values
                .get_mut(&handle)
                .ok_or_else(|| Self::invalid_handle_err(handle))?;
            handle_value.set_flags(flags);
            Ok(())
        }

        fn forget_handle(&mut self, handle: Handle) {
            self.handle_values.remove(&handle);
        }

        /// Increments the lookup count from the given path by 1.
        fn incr_lookup_count(&mut self, from: &Arc<BlockPath>) {
            match self.lookup_counts.entry(from.clone()) {
                hash_map::Entry::Occupied(mut entry) => {
                    // I don't want this to silently overflow.
                    let new_count = entry.get().checked_add(1).unwrap();
                    *entry.get_mut() = new_count;
                }
                hash_map::Entry::Vacant(entry) => {
                    entry.insert(1);
                }
            }
        }

        /// Decrements the lookup count from the given path by the given amount.
        fn decr_lookup_count(&mut self, from: &Arc<BlockPath>, decr: u64) {
            match self.lookup_counts.entry(from.clone()) {
                hash_map::Entry::Occupied(mut entry) => {
                    let new_count = entry.get().saturating_sub(decr);
                    if new_count > 0 {
                        *entry.get_mut() = new_count;
                    } else {
                        entry.remove();
                    }
                }
                hash_map::Entry::Vacant(..) => (),
            }
        }

        fn total_lookup_count(&self) -> u64 {
            self.lookup_counts.values().sum()
        }
    }

    type InodeTable<C> = HashMap<Inode, RwLock<InodeTableValue<C>>>;
    type InodeTableEntry<'a, C> = hash_map::Entry<'a, Inode, RwLock<InodeTableValue<C>>>;

    /// Structure for metadata about a blocktree.
    #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
    struct Superblock {
        /// The generation number of the cluster this part of the blocktree is stored on.
        generation: u64,
        /// The next free inode available to the cluster.
        next_inode: u64,
    }

    /// Structure for managing the part of a blocktree which is stored in the local filesystem.
    pub struct LocalFs<C, A> {
        /// The path to the directory in the local filesystem where this blocktree is located.
        path: PathBuf,
        /// A map from inode numbers to their reference counts.
        inodes: RwLock<InodeTable<C>>,
        /// The next inode that will be assigned to a new block.
        next_inode: AtomicU64,
        /// The generation number of this filesystem. This is the same for every other server in
        /// the same cluster.
        generation: u64,
        /// The credentials this blocktree instance will use for all cryptographic operations.
        creds: C,
        authorizer: A,
    }

    impl<C, A> LocalFs<C, A> {
        /// The maximum number of directory entries that can be returned in any given call to
        /// `read_dir`.
        const DIR_ENTRY_LIMIT: usize = 1024;
    }

    impl<C: Creds + 'static, A: Authorizer> LocalFs<C, A> {
        /// Creates a new empty blocktree at the given path.
        pub fn new_empty(
            btdir: PathBuf,
            generation: u64,
            creds: C,
            authorizer: A,
        ) -> Result<LocalFs<C, A>> {
            let mut root_block_path = creds
                .writecap()
                .ok_or(BlockError::MissingWritecap)?
                .root_block_path();

            // Initialize the superblock.
            let mut sb_block = Self::open_block(
                &btdir,
                SpecInodes::Sb.into(),
                creds.clone(),
                root_block_path.clone(),
            )?;
            let sb = Superblock {
                generation,
                next_inode: SpecInodes::FirstFree.into(),
            };
            write_to(&sb, &mut sb_block)?;
            sb_block.mut_meta_body().access_secrets(|secrets| {
                secrets.block_id.generation = generation;
                secrets.block_id.inode = SpecInodes::Sb.into();
                secrets.mode = FileType::Reg.value() | 0o666;
                secrets.uid = 0;
                secrets.gid = 0;
                secrets.nlink = 1;
                Ok(())
            })?;
            sb_block.flush()?;

            // Initialize the root directory.
            let mut root_block = Self::open_block(
                &btdir,
                SpecInodes::RootDir.into(),
                creds.clone(),
                root_block_path.clone(),
            )?;
            write_to(&Directory::new(), &mut root_block)?;
            root_block.mut_meta_body().access_secrets(|secrets| {
                secrets.block_id.generation = generation;
                secrets.block_id.inode = SpecInodes::RootDir.into();
                secrets.mode = FileType::Dir.value() | 0o777;
                secrets.uid = 0;
                secrets.gid = 0;
                secrets.nlink = 1;
                Ok(())
            })?;
            root_block.flush()?;

            let fs = Self::new(btdir, sb, sb_block, root_block, creds, authorizer)?;
            let writecap = fs.creds.writecap().ok_or(BlockError::MissingWritecap)?;
            let root_principal = writecap.root_principal();
            if fs.creds.principal() != root_principal {
                let proc_rec = IssuedProcRec {
                    addr: IpAddr::V6(Ipv6Addr::LOCALHOST),
                    pub_creds: fs.creds.concrete_pub(),
                    writecap: writecap.to_owned(),
                    authz_attrs: AuthzAttrs {
                        uid: 0,
                        gid: 0,
                        supp_gids: Vec::new(),
                    },
                };
                root_block_path.push_component(root_principal.to_string());
                fs.grant_access_to(
                    &Arc::new(root_block_path),
                    SpecInodes::RootDir.into(),
                    proc_rec,
                )?;
            }
            Ok(fs)
        }

        /// Opens an existing blocktree stored at the given path.
        pub fn new_existing(btdir: PathBuf, creds: C, authorizer: A) -> Result<LocalFs<C, A>> {
            let root_block_path = creds
                .writecap()
                .ok_or(BlockError::MissingWritecap)?
                .root_block_path();
            let mut sb_block = Self::open_block(
                &btdir,
                SpecInodes::Sb.into(),
                creds.clone(),
                root_block_path.to_owned(),
            )?;
            let sb = read_from(&mut sb_block)?;
            let root_block = Self::open_block(
                &btdir,
                SpecInodes::RootDir.into(),
                creds.clone(),
                root_block_path,
            )?;
            Self::new(btdir, sb, sb_block, root_block, creds, authorizer)
        }

        fn new(
            btdir: PathBuf,
            sb: Superblock,
            sb_block: Accessor<FileBlock<C>>,
            root_block: Accessor<FileBlock<C>>,
            creds: C,
            authorizer: A,
        ) -> Result<LocalFs<C, A>> {
            let mut inodes = HashMap::with_capacity(1);
            let empty_path = Arc::new(BlockPath::default());
            inodes.insert(
                SpecInodes::Sb.into(),
                RwLock::new(InodeTableValue::new(sb_block, empty_path.clone())),
            );
            inodes.insert(
                SpecInodes::RootDir.into(),
                RwLock::new(InodeTableValue::new(root_block, empty_path)),
            );
            Ok(LocalFs {
                path: btdir,
                inodes: RwLock::new(inodes),
                next_inode: AtomicU64::new(sb.next_inode),
                generation: sb.generation,
                creds,
                authorizer,
            })
        }

        /// Returns the path to the file storing the given inode's data.
        fn block_path<P: AsRef<Path>>(parent: P, inode: Inode) -> PathBuf {
            let group = inode / 0xFF;
            let mut path = PathBuf::new();
            path.push(parent);
            path.push(format!("{group:02x}"));
            path.push(format!("{inode:x}.blk"));
            path
        }

        fn open_block<P: AsRef<Path>>(
            btdir: P,
            inode: Inode,
            creds: C,
            block_path: BlockPath,
        ) -> Result<Accessor<FileBlock<C>>> {
            let path = Self::block_path(&btdir, inode);
            let dir = path.ancestors().nth(1).unwrap();
            if let Err(err) = std::fs::create_dir(dir) {
                match err.kind() {
                    io::ErrorKind::AlreadyExists => (),
                    _ => return Err(err.into()),
                }
            }
            let file = std::fs::OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .open(path)?;
            Self::open_block_file(file, creds, block_path)
        }

        fn open_block_file(
            file: File,
            creds: C,
            block_path: BlockPath,
        ) -> Result<Accessor<FileBlock<C>>> {
            let block = BlockOpenOptions::new()
                .with_creds(creds)
                .with_encrypt(true)
                .with_inner(file)
                .with_block_path(block_path)
                .open()?;
            Ok(block)
        }

        fn access_entry<T, F: FnOnce(InodeTableEntry<C>) -> Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> Result<T> {
            let mut inodes = self.inodes.write().display_err()?;
            let entry = inodes.entry(inode);
            cb(entry)
        }

        fn access_value<T, F: FnOnce(&InodeTableValue<C>) -> Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> Result<T> {
            let inodes = self.inodes.read().display_err()?;
            let guard = inodes
                .get(&inode)
                .ok_or_else(|| bterr!(Error::NotOpen(inode)))?
                .read()
                .display_err()?;
            cb(&guard)
        }

        fn access_value_mut<T, F: FnOnce(&mut InodeTableValue<C>) -> Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> Result<T> {
            let inodes = self.inodes.read().display_err()?;
            let mut guard = inodes
                .get(&inode)
                .ok_or_else(|| bterr!(Error::NotOpen(inode)))?
                .write()
                .display_err()?;
            cb(&mut guard)
        }

        fn access_block<T, F: FnOnce(&mut Accessor<&FileBlock<C>>, Flags) -> Result<T>>(
            &self,
            from: &BlockPath,
            inode: Inode,
            handle: Handle,
            cb: F,
        ) -> Result<T> {
            self.access_value(inode, |value| value.access_block(from, handle, cb))
        }

        fn access_block_mut<T, F: FnOnce(&mut Accessor<&mut FileBlock<C>>, Flags) -> Result<T>>(
            &self,
            from: &BlockPath,
            inode: Inode,
            handle: Handle,
            cb: F,
        ) -> Result<T> {
            self.access_value_mut(inode, |value| value.access_block_mut(from, handle, cb))
        }

        fn borrow_block<T, F: FnOnce(&Accessor<FileBlock<C>>) -> Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> Result<T> {
            self.access_value(inode, |value| value.borrow_block(cb))
        }

        fn borrow_block_mut<T, F: FnOnce(&mut Accessor<FileBlock<C>>) -> Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> Result<T> {
            self.access_value_mut(inode, |value| value.borrow_block_mut(cb))
        }

        fn open_value<T, F: FnOnce(&mut InodeTableValue<C>) -> Result<T>>(
            &self,
            from: Arc<BlockPath>,
            inode: Inode,
            block_path: BlockPath,
            cb: F,
        ) -> Result<T> {
            self.access_entry(inode, |entry| match entry {
                InodeTableEntry::Vacant(entry) => {
                    let block =
                        Self::open_block(&self.path, inode, self.creds.clone(), block_path)?;
                    let mut value = InodeTableValue::new(block, from);
                    let result = cb(&mut value);
                    entry.insert(RwLock::new(value));
                    result
                }
                InodeTableEntry::Occupied(mut entry) => {
                    let value = entry.get_mut().get_mut().display_err()?;
                    cb(value)
                }
            })
        }

        fn open_then_take_handle<T, F: FnOnce(Handle, &mut InodeTableValue<C>) -> Result<T>>(
            &self,
            from: Arc<BlockPath>,
            inode: Inode,
            block_path: BlockPath,
            flags: Flags,
            cb: F,
        ) -> Result<T> {
            self.open_value(from.clone(), inode, block_path, |value| {
                let handle = value.new_handle(from, flags)?;
                cb(handle, value)
            })
        }

        fn take_handle_if_ok<T, F: FnOnce(Handle, &mut InodeTableValue<C>) -> Result<T>>(
            &self,
            from: Arc<BlockPath>,
            inode: Inode,
            flags: Flags,
            cb: F,
        ) -> Result<T> {
            self.access_value_mut(inode, |value| {
                let handle = value.new_handle(from, flags)?;
                let result = cb(handle, value);
                if result.is_err() {
                    value.forget_handle(handle);
                }
                result
            })
        }

        fn inode_forget(
            &self,
            inodes: &mut RwLockWriteGuard<InodeTable<C>>,
            from: &Arc<BlockPath>,
            inode: Inode,
            count: u64,
        ) -> io::Result<()> {
            let lookup_count = {
                let inode_lock = match inodes.get_mut(&inode) {
                    Some(inode_lock) => inode_lock,
                    None => {
                        warn!("an attempt was made to forget non-existent inode {inode}");
                        return Ok(());
                    }
                };
                let mut value = inode_lock.write().display_err()?;
                value.decr_lookup_count(from, count);
                value.total_lookup_count()
            };
            if 0 == lookup_count {
                let delete = inodes
                    .remove(&inode)
                    .unwrap()
                    .into_inner()
                    .display_err()?
                    .delete;
                if delete {
                    let path = Self::block_path(&self.path, inode);
                    std::fs::remove_file(path)?;
                }
            }
            Ok(())
        }

        /// Returns the next available inode and updates the superblock in one atomic operation.
        /// TODO: Obviously this strategy won't work when there are multiple servers in this
        /// generation.
        fn next_inode(&self) -> Result<Inode> {
            self.borrow_block_mut(SpecInodes::Sb.into(), |mut block| {
                // We don't need strict ordering because the lock on the inode table value is already
                // serializing access.
                let inode = self.next_inode.fetch_add(1, Ordering::Relaxed);
                let sb = Superblock {
                    generation: self.generation,
                    next_inode: inode + 1,
                };
                block.rewind()?;
                write_to(&sb, &mut block)?;
                block.flush()?;
                Ok(inode)
            })
        }

        fn attr_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }

        fn entry_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }

        fn unsupported_flag_err(flag: &str) -> btlib::Error {
            bterr!("unsupported flag: {flag}")
        }

        fn bt_entry(&self, attr: BlockMetaSecrets) -> crate::msg::Entry {
            crate::msg::Entry {
                attr,
                attr_timeout: self.attr_timeout(),
                entry_timeout: self.entry_timeout(),
            }
        }

        /// Grants the given credentials access to the directory this instance is responsible for.
        ///
        /// # Warning
        /// This method calls `self.authz_attrs`, so the same consideration for avoiding deadlock
        /// apply to this method as well. See the documentation of `self.authz_attrs` for details.
        fn grant_access_to(
            &self,
            from: &Arc<BlockPath>,
            inode: Inode,
            proc_rec: IssuedProcRec,
        ) -> Result<()> {
            let authz_attrs = self.authz_attrs(from)?;
            let principal = proc_rec.pub_creds.principal();
            let next_inode = if inode == SpecInodes::RootDir.value() {
                // If the inode is for the root directory we need to add a readcap for the
                // superblock when we access it.
                self.borrow_block_mut(SpecInodes::Sb.into(), |mut block| {
                    let next_inode = self.next_inode.fetch_add(1, Ordering::Relaxed);
                    let sb = Superblock {
                        generation: self.generation,
                        next_inode: next_inode + 1,
                    };
                    block.rewind()?;
                    write_to(&sb, &mut block)?;
                    block
                        .mut_meta_body()
                        .add_readcap_for(principal.clone(), &proc_rec.pub_creds.enc)?;
                    block.flush()?;

                    Ok(next_inode)
                })
            } else {
                self.next_inode()
            }?;

            self.borrow_block_mut(inode, |block| {
                self.authorizer
                    .can_write(&AuthzContext::new(from, &authz_attrs, block.meta()))?;
                block
                    .mut_meta_body()
                    .add_readcap_for(principal.clone(), &proc_rec.pub_creds.enc)?;

                let mut dir = block.read_dir()?;
                let proc_rec_name = principal.to_string();
                dir.add_file(proc_rec_name, next_inode)?;
                // Note that write_dir calls flush, which also ensures metadata is written to disk.
                block.write_dir(&dir)?;

                Ok(())
            })?;

            let self_writecap = self.creds.writecap().ok_or(BlockError::MissingWritecap)?;
            let self_bind_path = Arc::new(self_writecap.bind_path());
            let bind_path = proc_rec.writecap.bind_path();
            self.open_value(self_bind_path.clone(), next_inode, bind_path, |value| {
                value.borrow_block_mut(|block| {
                    // TODO: This should not be needed because the principal has been given access
                    // to the root directory. Once proper inherited readcaps are implemented we will
                    // no longer need to add a readcap to this block.
                    block
                        .mut_meta_body()
                        .add_readcap_for(principal, &proc_rec.pub_creds.enc)?;
                    block.write_proc_rec(&ProcRec::Valid(proc_rec))
                })
            })?;
            // We must ensure the reference count for the inode is decremented, otherwise the table
            // entry will never be freed.
            let mut inodes = self.inodes.write().display_err()?;
            self.inode_forget(&mut inodes, &self_bind_path, next_inode, 1)?;
            Ok(())
        }

        fn lookup_inode_in(&self, parent: Inode, name: &str) -> Result<Inode> {
            let dir = self.borrow_block_mut(parent, |block| block.read_dir())?;
            dir.entry(name)
                .ok_or_else(|| io::Error::from_raw_os_error(libc::ENOENT).into())
                .map(|e| e.inode())
        }

        fn lookup_inode<'a, I: Iterator<Item = &'a str>>(&self, components: I) -> Result<Inode> {
            const ROOT: Inode = SpecInodes::RootDir as Inode;
            let mut inode = ROOT;
            for component in components {
                inode = self.lookup_inode_in(inode, component)?;
            }
            if inode == ROOT {
                Err(io::Error::from_raw_os_error(libc::ENOENT).into())
            } else {
                Ok(inode)
            }
        }

        /// Retrieves the authorization attributes for the principal identified by the given path.
        /// If the principal is not associated with a valid process record, then an [Err] is
        /// returned.
        /// ## Warning
        /// If this method is called while a lock for any component on the given path is held, then
        /// a deadlock may occur. It's safest to call this method when _no_ locks are held.
        fn authz_attrs(&self, from: &Arc<BlockPath>) -> Result<AuthzAttrs> {
            let writecap = self.creds.writecap().ok_or(BlockError::MissingWritecap)?;
            let root_principal = writecap.root_principal();
            let from_principal = from.components().last().map_or_else(
                || Err(bterr!("path from which message was received was empty")),
                Principal::try_from,
            )?;
            if root_principal == from_principal {
                // Now I am become root, the destroyer of files.
                return Ok(AuthzAttrs {
                    uid: 0,
                    gid: 0,
                    supp_gids: Vec::new(),
                });
            }
            let local_root = writecap.path();
            let relative = from.relative_to(local_root)?;
            let inode = self.lookup_inode(relative.components())?;
            let proc_rec =
                self.open_value(from.clone(), inode, from.as_ref().to_owned(), |value| {
                    value.borrow_block_mut(|block| block.read_proc_rec())
                })?;
            let proc_rec = proc_rec.validate()?;
            Ok(proc_rec.authz_attrs)
        }
    }

    unsafe impl<C: Sync, A: Sync> Sync for LocalFs<C, A> {}

    impl<C: 'static + Creds + Clone + Send + Sync, A: 'static + Authorizer + Send + Sync> FsProvider
        for LocalFs<C, A>
    {
        type LookupFut<'c> = Ready<Result<LookupReply>>;
        fn lookup<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Lookup<'c>) -> Self::LookupFut<'c> {
            let result = (move || {
                let Lookup { parent, name, .. } = msg;
                debug!("lookup: parent {parent}, {:?}", name);
                let authz_attrs = self.authz_attrs(from)?;

                let (dir, block_path) = match self.borrow_block_mut(parent, |parent_block| {
                    self.authorizer.can_exec(&AuthzContext::new(
                        from,
                        &authz_attrs,
                        parent_block.meta(),
                    ))?;
                    let dir = parent_block.read_dir()?;
                    let path = parent_block.meta_body().path().to_owned();
                    Ok((dir, path))
                }) {
                    Ok(pair) => pair,
                    Err(err) => {
                        error!("lookup failed to borrow inode {parent}: {err}");
                        return Err(err);
                    }
                };
                let entry = dir
                    .entry(name)
                    .ok_or_else(|| io::Error::from_raw_os_error(libc::ENOENT))?;
                let inode = entry.inode();
                let stat = match self.open_value(from.to_owned(), inode, block_path, |value| {
                    let stat = value.block.meta_body().secrets()?.to_owned();
                    value.incr_lookup_count(from);
                    Ok(stat)
                }) {
                    Ok(stat) => stat,
                    Err(err) => {
                        error!("lookup failed to read stats for '{name}': {err}");
                        return Err(err);
                    }
                };
                let entry = self.bt_entry(stat);
                let reply = LookupReply {
                    inode,
                    generation: self.generation,
                    entry,
                };
                Ok(reply)
            })();
            ready(result)
        }

        type CreateFut<'c> = Ready<Result<CreateReply>>;
        fn create<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Create<'c>) -> Self::CreateFut<'c> {
            let result = (move || {
                let Create {
                    parent,
                    name,
                    flags,
                    mode,
                    umask,
                } = msg;
                debug!("create: parent {parent}, name {:?}", name);
                let authz_attrs = self.authz_attrs(from)?;

                let name = msg.name.to_owned();

                // Add a directory entry to the parent for the new inode.
                let (inode, mut block_path) = self.borrow_block_mut(parent, |block| {
                    self.authorizer.can_write(&AuthzContext::new(
                        from,
                        &authz_attrs,
                        block.meta(),
                    ))?;

                    let mut dir = block.read_dir()?;
                    if dir.contains_entry(&name) {
                        return Err(io::Error::from_raw_os_error(libc::EEXIST).into());
                    }

                    // Reserve a free inode.
                    let inode = self.next_inode()?;
                    dir.add_file(name.clone(), inode)?;
                    block.write_dir(&dir)?;

                    Ok((inode, block.meta_body().path().clone()))
                })?;
                block_path.push_component(name);

                let (handle, stat) = self.open_then_take_handle(
                    from.to_owned(),
                    inode,
                    block_path,
                    libc::O_RDWR.into(),
                    |handle, value| {
                        let stat = value.access_block_mut(from, handle, |block, _| {
                            let stat = block.mut_meta_body().access_secrets(|secrets| {
                                secrets.block_id.generation = self.generation;
                                secrets.block_id.inode = inode;
                                secrets.mode = mode & !umask;
                                if flags.directory() {
                                    secrets.mode |= FileType::Dir.value();
                                } else {
                                    secrets.mode |= FileType::Reg.value();
                                }
                                secrets.uid = authz_attrs.uid;
                                secrets.gid = authz_attrs.gid;
                                let now = Epoch::now();
                                secrets.atime = now;
                                secrets.ctime = now;
                                secrets.mtime = now;
                                secrets.nlink = 1;
                                Ok(secrets.to_owned())
                            })?;
                            if flags.directory() {
                                // Note that write_dir flushes data after writing, including
                                // metadata.
                                block.write_dir(&Directory::new())?;
                            } else {
                                block.flush_meta()?;
                            }
                            Ok(stat)
                        })?;
                        if flags.directory() {
                            value.convert_to_dir(handle)?;
                        }
                        value.set_flags(handle, flags)?;
                        Ok((handle, stat))
                    },
                )?;

                Ok(CreateReply {
                    inode,
                    handle,
                    entry: self.bt_entry(stat),
                })
            })();
            ready(result)
        }

        type OpenFut<'c> = Ready<Result<OpenReply>>;
        fn open<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Open) -> Self::OpenFut<'c> {
            let result = (move || {
                let Open { inode, flags } = msg;
                debug!("open: inode {inode}, flags {flags}");
                if flags.value() & libc::O_APPEND != 0 {
                    return Err(Self::unsupported_flag_err("O_APPEND"));
                }
                if flags.value() & libc::O_CLOEXEC != 0 {
                    return Err(Self::unsupported_flag_err("O_CLOEXEC"));
                }
                let authz_attrs = self.authz_attrs(from)?;
                let handle =
                    self.take_handle_if_ok(from.clone(), inode, flags, |handle, value| {
                        value.access_block(from, handle, |block, _| {
                            let ctx = AuthzContext::new(from, &authz_attrs, block.meta());
                            if flags.readable() {
                                self.authorizer.can_read(&ctx)?;
                            }
                            if flags.writeable() {
                                self.authorizer.can_write(&ctx)?;
                            }
                            Ok(())
                        })?;
                        if flags.directory() {
                            value.convert_to_dir(handle)?;
                        }
                        value.set_flags(handle, flags)?;
                        Ok(handle)
                    })?;
                Ok(OpenReply { handle })
            })();
            ready(result)
        }

        fn read<'c, R, F>(&'c self, from: &'c Arc<BlockPath>, msg: Read, callback: F) -> Result<R>
        where
            F: 'c + FnOnce(&[u8]) -> R,
        {
            let Read {
                inode,
                handle,
                offset,
                size,
            } = msg;
            debug!("read: inode {inode}, handle {handle}, offset {offset}, size {size}");
            let mut callback = Some(callback);
            let output = self.access_block(from, inode, handle, |block, flags| {
                flags.assert_readable()?;
                let pos = block.pos() as u64;
                if offset != pos {
                    if let Err(err) = block.try_seek(SeekFrom::Start(offset)) {
                        //  An error with `ErrorKind::Unsupported` means that the `SectoredBuf`
                        // has unflushed data and it needs exclusive access to the block to
                        // perform this seek because this data needs to be written.
                        if let io::ErrorKind::Unsupported = err.kind() {
                            return Ok(None);
                        } else {
                            return Err(err.into());
                        }
                    }
                }
                let buf = block.get_buf(offset, size)?;
                let callback = callback.take().unwrap();
                Ok(Some(callback(buf)))
            })?;
            let output = match output {
                Some(output) => output,
                None => {
                    // The offset of this read requires us to flush data buffered from a previous
                    // write before seeking to a different sector, so we have to access the block
                    // mutably.
                    self.access_block_mut(from, inode, handle, |block, _| {
                        block.seek(SeekFrom::Start(offset))?;
                        let buf = block.get_buf(offset, size)?;
                        let callback = callback.take().unwrap();
                        Ok(callback(buf))
                    })?
                }
            };
            Ok(output)
        }

        type WriteFut<'c> = Ready<Result<WriteReply>>;
        fn write<'c, R>(
            &'c self,
            from: &'c Arc<BlockPath>,
            inode: Inode,
            handle: Handle,
            offset: u64,
            size: u64,
            reader: R,
        ) -> Self::WriteFut<'c>
        where
            R: 'c + IoRead,
        {
            let result = (move || {
                debug!("write: inode {inode}, handle {handle}, offset {offset}");
                let written = self.access_block_mut(from, inode, handle, |block, flags| {
                    flags.assert_writeable()?;
                    let pos = block.pos() as u64;
                    if offset != pos {
                        block.seek(SeekFrom::Start(offset))?;
                    }
                    let size = size.try_into().display_err()?;
                    block.write_from(reader, size).bterr()
                })?;
                Ok(WriteReply {
                    written: written as u64,
                })
            })();
            ready(result)
        }

        type FlushFut<'c> = Ready<Result<()>>;
        fn flush<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Flush) -> Self::FlushFut<'c> {
            let result = (|| {
                let Flush { inode, handle } = msg;
                debug!("flush: inode {inode}, handle {handle}");
                match self.access_block_mut(from, inode, handle, |block, flags| {
                    flags.assert_writeable()?;
                    block.flush().bterr()
                }) {
                    Ok(value) => Ok(value),
                    Err(err) => match err.downcast_ref::<Error>() {
                        // If the handle is read-only we just return Ok.
                        Some(Error::ReadOnlyHandle(..)) => Ok(()),
                        _ => Err(err),
                    },
                }
            })();
            ready(result)
        }

        type ReadDirFut<'c> = Ready<Result<ReadDirReply>>;
        fn read_dir<'c>(&'c self, from: &'c Arc<BlockPath>, msg: ReadDir) -> Self::ReadDirFut<'c> {
            let result = (|| {
                let ReadDir {
                    inode,
                    handle,
                    limit,
                    state,
                } = msg;
                debug!("read_dir: inode {inode}, handle {handle}, state {state}");
                self.access_value(inode, |value| {
                    let handle_value = value
                        .value(handle)
                        .map_err(|_| bterr!(Error::InvalidHandle { handle, inode }))?;
                    let (_, owner, flags) = handle_value.parts();
                    flags.assert_readable()?;
                    owner.assert_eq(from)?;
                    let dir = handle_value.directory()?;
                    let state: usize = state.try_into()?;
                    let server_limit = Self::DIR_ENTRY_LIMIT.min(dir.num_entries() - state);
                    let entries_len = if limit > 0 {
                        server_limit.min(limit as usize)
                    } else {
                        server_limit
                    };
                    let pairs = dir
                        .entries()
                        .skip(state)
                        .take(entries_len)
                        .map(|(name, entry)| (name.to_owned(), entry.to_owned()));
                    let mut entries = Vec::with_capacity(entries_len);
                    entries.extend(pairs);
                    let reply = ReadDirReply {
                        entries,
                        new_state: (state + entries_len) as u64,
                    };
                    Ok(reply)
                })
            })();
            ready(result)
        }

        type LinkFut<'c> = Ready<Result<LinkReply>>;
        fn link<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Link) -> Self::LinkFut<'c> {
            let result = (move || {
                let Link {
                    inode,
                    new_parent,
                    name,
                } = msg;
                debug!("link: inode {inode}, new_parent {new_parent}, name {name}");
                let authz_attrs = self.authz_attrs(from)?;
                self.borrow_block_mut(new_parent, |parent_block| {
                    self.authorizer.can_write(&AuthzContext::new(
                        from,
                        &authz_attrs,
                        parent_block.meta(),
                    ))?;

                    let mut dir = parent_block.read_dir()?;
                    if dir.contains_entry(name) {
                        return Err(io::Error::from_raw_os_error(libc::EEXIST).into());
                    }

                    let attr = self.access_value_mut(inode, |value| {
                        let block = value.block_mut();
                        let meta = block.mut_meta_body();
                        let attr = meta.access_secrets(|secrets| {
                            secrets.nlink += 1;
                            Ok(secrets.to_owned())
                        })?;
                        block.flush_meta()?;
                        value.incr_lookup_count(from);
                        Ok(attr)
                    })?;
                    let file_type = FileType::from_value(attr.mode)?;
                    let entry = match file_type {
                        FileType::Reg => DirEntry::File(inode),
                        FileType::Dir => DirEntry::Directory(inode),
                    };
                    dir.insert_entry(name.to_owned(), entry);
                    parent_block.write_dir(&dir)?;
                    let entry = self.bt_entry(attr);
                    Ok(LinkReply { entry })
                })
            })();
            ready(result)
        }

        type UnlinkFut<'c> = Ready<Result<()>>;
        fn unlink<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Unlink) -> Self::UnlinkFut<'c> {
            let result = (move || {
                let Unlink { parent, name } = msg;
                debug!("unlink: parent {parent}, name {name}");
                let authz_attrs = self.authz_attrs(from)?;
                let (block_path, inode) = self.borrow_block_mut(parent, |parent_block| {
                    self.authorizer.can_write(&AuthzContext::new(
                        from,
                        &authz_attrs,
                        parent_block.meta(),
                    ))?;

                    let mut dir = parent_block.read_dir()?;
                    let entry = match dir.remove_entry(name) {
                        None => return Err(io::Error::from_raw_os_error(libc::ENOENT).into()),
                        Some(entry) => entry,
                    };
                    let inode = entry.inode();
                    parent_block.write_dir(&dir)?;

                    let mut block_path = parent_block.meta_body().path().clone();
                    block_path.push_component(name.to_owned());
                    Ok((block_path, inode))
                })?;
                self.open_value(from.to_owned(), inode, block_path, |value| {
                    // We mark the block for deletion if `nlink` drops to zero.
                    let block = value.block_mut();
                    let nlink = block.mut_meta_body().access_secrets(|secrets| {
                        secrets.nlink -= 1;
                        Ok(secrets.nlink)
                    })?;
                    block.flush_meta()?;
                    value.delete = 0 == nlink;
                    Ok(())
                })
            })();
            ready(result)
        }

        type ReadMetaFut<'c> = Ready<Result<ReadMetaReply>>;
        fn read_meta<'c>(
            &'c self,
            from: &'c Arc<BlockPath>,
            msg: ReadMeta,
        ) -> Self::ReadMetaFut<'c> {
            let result = (move || {
                let ReadMeta { inode, handle } = msg;
                debug!("read_meta: inode {inode}, handle {:?}", handle);
                let attrs = if let Some(handle) = handle {
                    self.access_block(from, inode, handle, |block, _| {
                        Ok(block.meta_body().secrets()?.to_owned())
                    })?
                } else {
                    self.borrow_block(inode, |block| Ok(block.meta_body().secrets()?.to_owned()))?
                };
                debug!("read_meta attrs: {:?}", attrs);
                let reply = ReadMetaReply {
                    attrs,
                    valid_for: self.attr_timeout(),
                };
                Ok(reply)
            })();
            ready(result)
        }

        type WriteMetaFut<'c> = Ready<Result<WriteMetaReply>>;
        fn write_meta<'c>(
            &'c self,
            from: &'c Arc<BlockPath>,
            msg: WriteMeta,
        ) -> Self::WriteMetaFut<'c> {
            let result = (move || {
                let WriteMeta {
                    inode,
                    handle,
                    attrs,
                    attrs_set,
                } = msg;
                debug!("write_meta: inode {inode}, handle {:?}", handle);
                let authz_attrs = self.authz_attrs(from)?;
                let cb = |block: &mut FileBlock<C>| {
                    self.authorizer.can_write(&AuthzContext::new(
                        from,
                        &authz_attrs,
                        block.meta(),
                    ))?;
                    let attrs = block.mut_meta_body().access_secrets(|secrets| {
                        if attrs_set.mode() {
                            secrets.mode = attrs.mode;
                        }
                        if attrs_set.uid() {
                            secrets.uid = attrs.uid;
                        }
                        if attrs_set.gid() {
                            secrets.gid = attrs.gid;
                        }
                        if attrs_set.atime() {
                            secrets.atime = attrs.atime;
                        }
                        if attrs_set.mtime() {
                            secrets.mtime = attrs.mtime;
                        }
                        if attrs_set.ctime() {
                            secrets.ctime = attrs.ctime;
                        }
                        for (key, value) in attrs.tags.into_iter() {
                            secrets.tags.insert(key, value);
                        }
                        Ok(secrets.to_owned())
                    })?;
                    block.flush_meta()?;
                    Ok(attrs)
                };
                let attrs = if let Some(handle) = handle {
                    self.access_block_mut(from, inode, handle, |block, flags| {
                        flags.assert_writeable()?;
                        cb(block.get_mut())
                    })
                } else {
                    self.borrow_block_mut(inode, |block| cb(block.get_mut()))
                }?;
                Ok(WriteMetaReply {
                    attrs,
                    valid_for: self.attr_timeout(),
                })
            })();
            ready(result)
        }

        type AllocateFut<'c> = Ready<Result<()>>;
        fn allocate<'c>(
            &'c self,
            from: &'c Arc<BlockPath>,
            msg: Allocate,
        ) -> Self::AllocateFut<'c> {
            let Allocate {
                inode,
                handle,
                offset,
                size,
            } = msg;
            debug!(
                "allocate: inode {inode}, handle {handle}, offset {:?}, size {size}",
                offset
            );
            let result = self.access_block_mut(from, inode, handle, |block, _| {
                let curr_size = block.meta_body().secrets()?.size;
                if let Some(offset) = offset {
                    if curr_size != offset {
                        return Err(bterr!("only allocations at the end of files are supported"));
                    }
                }
                let new_size = curr_size.max(size);
                if new_size > curr_size {
                    block.zero_extend(new_size - curr_size)?;
                }
                Ok(())
            });
            ready(result)
        }

        type CloseFut<'c> = Ready<Result<()>>;
        fn close<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Close) -> Self::CloseFut<'c> {
            let Close { inode, handle } = msg;
            debug!("close: inode {inode}, handle {handle}");
            let result = self.access_value_mut(inode, |value| {
                if let Err(err) = value.access_block_mut(from, handle, |block, _| {
                    block.flush()?;
                    block.flush_meta()?;
                    Ok(())
                }) {
                    if let Some(Error::ReadOnlyHandle(_)) = err.downcast_ref::<Error>() {
                    } else {
                        return Err(err);
                    }
                }
                value.forget_handle(handle);
                Ok(())
            });
            ready(result)
        }

        type ForgetFut<'c> = Ready<Result<()>>;
        fn forget<'c>(&'c self, from: &'c Arc<BlockPath>, msg: Forget) -> Self::ForgetFut<'c> {
            let result = (move || {
                let Forget { inode, count } = msg;
                debug!("forget: inode {inode}, count {count}");
                let mut inodes = self.inodes.write().display_err()?;
                self.inode_forget(&mut inodes, from, inode, count).bterr()
            })();
            ready(result)
        }

        type LockFut<'c> = Ready<Result<()>>;
        fn lock<'c>(&'c self, _from: &'c Arc<BlockPath>, _msg: Lock) -> Self::LockFut<'c> {
            todo!();
        }

        type UnlockFut<'c> = Ready<Result<()>>;
        fn unlock<'c>(&'c self, _from: &'c Arc<BlockPath>, _msg: Unlock) -> Self::UnlockFut<'c> {
            todo!();
        }

        type AddReacapFut<'c> = Ready<Result<()>>;
        fn add_readcap<'c>(
            &'c self,
            from: &'c Arc<BlockPath>,
            msg: AddReadcap,
        ) -> Self::AddReacapFut<'c> {
            let AddReadcap {
                inode,
                handle,
                principal,
                enc_key,
            } = msg;
            debug!("add_readcap: inode {inode}, handle {handle}, principal {principal}");
            let result = self.access_block_mut(from, inode, handle, |block, _| {
                block.mut_meta_body().add_readcap_for(principal, &enc_key)
            });
            ready(result)
        }

        type GrantAccessFut<'c> = Ready<Result<()>>;
        fn grant_access<'c>(
            &'c self,
            from: &'c Arc<BlockPath>,
            msg: GrantAccess,
        ) -> Self::GrantAccessFut<'c> {
            let GrantAccess { inode, record } = msg;
            debug!("grant_access: inode {inode}, record {:?}", record);
            ready(self.grant_access_to(from, inode, record))
        }
    }
}
