// SPDX-License-Identifier: AGPL-3.0-or-later
mod fuse_daemon;
use fuse_daemon::FuseDaemon;
mod config;
mod fuse_fs;

use config::{Config, ConfigRef, EnvVars};

use btfproto::{local_fs::LocalFs, server::FsProvider};
use btlib::{
    config_helpers::from_envvar,
    crypto::{
        tpm::{TpmCredStore, TpmCreds},
        CredStore, Creds, CredsPriv,
    },
    Result,
};
use core::future::Future;
use std::{
    fs::{self},
    io,
    path::{Path, PathBuf},
    sync::Arc,
};

const ENVVARS: EnvVars = EnvVars {
    tabrmd: "BT_TABRMD",
    mnt_options: "BT_MNTOPTS",
};

const DEFAULT_CONFIG: ConfigRef<'static> = ConfigRef {
    block_dir: "./bt",
    mnt_dir: "./mnt",
    tpm_state_file: "./tpm_state",
    tabrmd: "bus_type=session",
    mnt_options: "default_permissions",
    threads: 8,
};

trait PathExt {
    fn try_create_dir(&self) -> io::Result<()>;
}

impl<T: AsRef<Path>> PathExt for T {
    fn try_create_dir(&self) -> io::Result<()> {
        match fs::create_dir(self) {
            Ok(_) => Ok(()),
            Err(err) => match err.kind() {
                io::ErrorKind::AlreadyExists => Ok(()),
                _ => Err(err),
            },
        }
    }
}

fn node_creds(state_file: PathBuf, tabrmd_cfg: &str) -> Result<TpmCreds> {
    let cred_store = TpmCredStore::from_tabrmd(tabrmd_cfg, state_file)?;
    cred_store.node_creds()
}

fn provider<C: 'static + Creds + Send + Sync>(
    btdir: PathBuf,
    node_creds: C,
) -> Result<impl FsProvider> {
    btdir.try_create_dir()?;
    let empty = fs::read_dir(&btdir)?.next().is_none();
    if empty {
        LocalFs::new_empty(btdir, 0, node_creds, btfproto::local_fs::ModeAuthorizer {})
    } else {
        LocalFs::new_existing(btdir, node_creds, btfproto::local_fs::ModeAuthorizer {})
    }
}

fn run_daemon(config: Config) -> impl Send + Sync + Future<Output = ()> {
    let node_creds =
        node_creds(config.tpm_state_file, &config.tabrmd).expect("failed to get node creds");
    let fallback_path = {
        let writecap = node_creds
            .writecap()
            .ok_or(btlib::BlockError::MissingWritecap)
            .unwrap();
        Arc::new(writecap.bind_path())
    };
    let provider = provider(config.block_dir, node_creds).expect("failed to create FS provider");

    let mut daemon = FuseDaemon::new(config.mnt_dir, config.threads, fallback_path, provider)
        .expect("failed to create FUSE daemon");
    async move { daemon.finished().await }
}

#[tokio::main]
async fn main() {
    env_logger::init();
    let mut args = std::env::args_os().skip(1).map(PathBuf::from);
    let builder = Config::builder()
        .with_block_dir(args.next())
        .with_mnt_dir(args.next())
        .with_tabrmd(from_envvar(ENVVARS.tabrmd).unwrap())
        .with_mnt_options(from_envvar(ENVVARS.mnt_options).unwrap());
    let config = builder.build();
    run_daemon(config).await;
}

#[cfg(test)]
mod test {
    use super::*;

    use btlib::{crypto::Creds, log::BuilderExt, Epoch, Principaled};
    use ctor::ctor;
    use std::{
        ffi::{OsStr, OsString},
        fs::{
            create_dir, hard_link, metadata, read, read_dir, remove_dir, remove_file, rename,
            set_permissions, write, Permissions, ReadDir,
        },
        os::unix::fs::PermissionsExt,
        sync::mpsc,
        thread::JoinHandle,
        time::Duration,
    };
    use swtpm_harness::SwtpmHarness;
    use tempdir::TempDir;

    /// An optional timeout to wait for the FUSE daemon to start in tests.
    const TIMEOUT: Option<Duration> = Some(Duration::from_millis(1000));

    /// The log level to use when running tests.
    /// Note that the debug log level significantly reduces performance.
    const LOG_LEVEL: &str = "warn";

    #[ctor]
    fn ctor() {
        std::env::set_var("RUST_LOG", LOG_LEVEL);
        env_logger::Builder::from_default_env().btformat().init();
    }

    /// Unmounts the file system at the given path.
    fn unmount<P: AsRef<Path>>(mnt_path: P) {
        const PROG: &str = "fusermount";
        let mnt_path = mnt_path
            .as_ref()
            .as_os_str()
            .to_str()
            .expect("failed to convert mnt_path to `str`");
        let code = std::process::Command::new(PROG)
            .args(["-z", "-u", mnt_path])
            .status()
            .expect("waiting for exit status failed")
            .code()
            .expect("code returned None");
        if code != 0 {
            panic!("{PROG} exited with a non-zero status: {code}");
        }
    }

    fn file_names(read_dir: ReadDir) -> impl Iterator<Item = OsString> {
        read_dir.map(|entry| entry.unwrap().file_name())
    }

    const ROOT_PASSWD: &str = "password";

    struct TestCase {
        config: Config,
        handle: Option<JoinHandle<()>>,
        node_principal: OsString,
        // Note that the drop order of these fields is significant.
        _cred_store: TpmCredStore,
        _swtpm: SwtpmHarness,
        _temp_dir: TempDir,
    }

    impl TestCase {
        fn new() -> TestCase {
            let tmp = TempDir::new("btfuse").unwrap();
            let (mounted_tx, mounted_rx) = mpsc::channel();
            let (swtpm, cred_store) = Self::swtpm();
            let config = Config::builder()
                .with_block_dir(Some(tmp.path().join("bt")))
                .with_mnt_dir(Some(tmp.path().join("mnt")))
                .with_tpm_state_file(Some(swtpm.state_path().to_owned().into()))
                .with_tabrmd(Some(swtpm.tabrmd_config().to_owned()))
                .build();
            let config_clone = config.clone();
            let handle = std::thread::spawn(move || Self::run(mounted_tx, config_clone));
            match TIMEOUT {
                Some(duration) => mounted_rx.recv_timeout(duration).unwrap(),
                None => mounted_rx.recv().unwrap(),
            };
            let node_principal =
                OsString::from(cred_store.node_creds().unwrap().principal().to_string());
            Self {
                config,
                handle: Some(handle),
                node_principal,
                _temp_dir: tmp,
                _swtpm: swtpm,
                _cred_store: cred_store,
            }
        }

        fn run(mounted_tx: mpsc::Sender<()>, config: Config) {
            let runtime = tokio::runtime::Builder::new_current_thread()
                .build()
                .unwrap();
            // run_daemon can only be called in the context of a runtime, hence the need to call
            // spawn_blocking.
            let started = runtime.spawn_blocking(|| run_daemon(config));
            let future = runtime.block_on(started).unwrap();

            // The file system is mounted before run_daemon returns.
            mounted_tx.send(()).unwrap();
            runtime.block_on(future);
        }

        fn swtpm() -> (SwtpmHarness, TpmCredStore) {
            let swtpm = SwtpmHarness::new().unwrap();
            let state_path: PathBuf = swtpm.state_path().to_owned();
            let cred_store = {
                let context = swtpm.context().unwrap();
                TpmCredStore::from_context(context, state_path.clone()).unwrap()
            };
            let root_creds = cred_store.gen_root_creds(ROOT_PASSWD).unwrap();
            let mut node_creds = cred_store.node_creds().unwrap();
            let expires = Epoch::now() + Duration::from_secs(3600);
            let writecap = root_creds
                .issue_writecap(node_creds.principal(), vec![], expires)
                .unwrap();
            cred_store
                .assign_node_writecap(&mut node_creds, writecap)
                .unwrap();
            (swtpm, cred_store)
        }

        fn mnt_dir(&self) -> &PathBuf {
            &self.config.mnt_dir
        }

        fn wait(&mut self) {
            if let Some(handle) = self.handle.take() {
                handle.join().expect("join failed");
            }
        }

        fn unmount_and_wait(&mut self) {
            // If `handle` has already been taken that means `wait` has already been called. If
            // this thread called `wait` and subsequently became unblocked, we know that the FUSE
            // thread halted, which only happens when the file system is unmounted. Hence
            // we don't have to unmount it, nor wait for the FUSE thread.
            if self.handle.is_none() {
                return;
            }
            let mnt_path = self.mnt_dir();
            unmount(mnt_path);
            self.wait();
        }

        fn initial_contents(&self) -> Vec<&OsStr> {
            vec![&self.node_principal]
        }
    }

    impl Drop for TestCase {
        fn drop(&mut self) {
            self.unmount_and_wait()
        }
    }

    /// Creates a new file system and mounts it at `/tmp/btfuse.<random>/mnt` so it can be
    /// tested manually.
    //#[test]
    #[allow(dead_code)]
    fn manual_test() {
        let mut case = TestCase::new();
        case.wait()
    }

    #[test]
    fn write_read() -> Result<()> {
        const EXPECTED: &[u8] =
            b"The paths to failure are uncountable, yet to success there is but one.";
        let case = TestCase::new();
        let file_path = case.mnt_dir().join("file");

        write(&file_path, EXPECTED)?;

        let actual = read(&file_path)?;
        assert_eq!(EXPECTED, actual);
        Ok(())
    }

    #[test]
    fn create_file_then_readdir() {
        const DATA: &[u8] = b"Au revoir Shoshanna!";
        let file_name = OsStr::new("landa_dialog.txt");
        let case = TestCase::new();
        let mut expected = case.initial_contents();
        expected.push(file_name);
        let mnt_path = case.mnt_dir();
        let file_path = mnt_path.join(file_name);

        write(&file_path, DATA).expect("write failed");

        let first = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(first.eq(expected.iter().map(|e| *e)));
        let second = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(second.eq(expected));
    }

    #[test]
    fn create_then_delete_file() {
        const DATA: &[u8] = b"The universe is hostile, so impersonal. Devour to survive";
        let file_name = OsStr::new("tool_lyrics.txt");
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let file_path = mnt_path.join(file_name);
        write(&file_path, DATA).expect("write failed");

        remove_file(&file_path).expect("remove_file failed");

        let expected = case.initial_contents();
        let actual = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(actual.eq(expected));
    }

    #[test]
    fn hard_link_then_remove() {
        const EXPECTED: &[u8] = b"And the lives we've reclaimed";
        let name1 = OsStr::new("refugee_lyrics.txt");
        let name2 = OsStr::new("rise_against_lyrics.txt");
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let path1 = mnt_path.join(name1);
        let path2 = mnt_path.join(name2);
        write(&path1, EXPECTED).expect("write failed");

        hard_link(&path1, &path2).expect("hard_link failed");
        remove_file(&path1).expect("remove_file failed");

        let actual = read(&path2).expect("read failed");
        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn hard_link_then_remove_both() {
        const EXPECTED: &[u8] = b"And the lives we've reclaimed";
        let name1 = OsStr::new("refugee_lyrics.txt");
        let name2 = OsStr::new("rise_against_lyrics.txt");
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let path1 = mnt_path.join(name1);
        let path2 = mnt_path.join(name2);
        write(&path1, EXPECTED).expect("write failed");

        hard_link(&path1, &path2).expect("hard_link failed");
        remove_file(&path1).expect("remove_file on path1 failed");
        remove_file(&path2).expect("remove_file on path2 failed");

        let expected = case.initial_contents();
        assert!(file_names(read_dir(&mnt_path).expect("read_dir failed")).eq(expected));
    }

    #[test]
    fn set_mode_bits() {
        const EXPECTED: u32 = libc::S_IFREG | 0o777;
        let case = TestCase::new();
        let file_path = case.mnt_dir().join("bagobits");
        write(&file_path, []).expect("write failed");
        let original = metadata(&file_path)
            .expect("metadata failed")
            .permissions()
            .mode();
        assert_ne!(EXPECTED, original);

        set_permissions(&file_path, Permissions::from_mode(EXPECTED))
            .expect("set_permissions failed");

        let actual = metadata(&file_path)
            .expect("metadata failed")
            .permissions()
            .mode();
        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn create_directory() {
        const EXPECTED: &str = "etc";
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let dir_path = mnt_path.join(EXPECTED);
        let mut expected = case.initial_contents();
        expected.push(OsStr::new(EXPECTED));

        create_dir(&dir_path).expect("create_dir failed");

        let actual = file_names(read_dir(mnt_path).expect("read_dir failed"));
        assert!(actual.eq(expected));
    }

    #[test]
    fn create_file_under_new_directory() {
        const DIR_NAME: &str = "etc";
        const FILE_NAME: &str = "file";
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let dir_path = mnt_path.join(DIR_NAME);
        let file_path = dir_path.join(FILE_NAME);

        create_dir(&dir_path).expect("create_dir failed");
        write(&file_path, []).expect("write failed");

        let actual = file_names(read_dir(dir_path).expect("read_dir failed"));
        assert!(actual.eq([FILE_NAME]));
    }

    #[test]
    fn create_then_remove_directory() {
        const DIR_NAME: &str = "etc";
        let case = TestCase::new();
        let mnt_path = case.mnt_dir();
        let dir_path = mnt_path.join(DIR_NAME);

        create_dir(&dir_path).expect("create_dir failed");
        remove_dir(&dir_path).expect("remove_dir failed");

        let actual = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(actual.eq(case.initial_contents()));
    }

    #[test]
    fn read_only_dir_cant_create_subdir() {
        const DIR_NAME: &str = "etc";
        let case = TestCase::new();
        let dir_path = case.mnt_dir().join(DIR_NAME);
        create_dir(&dir_path).expect("create_dir failed");
        set_permissions(&dir_path, Permissions::from_mode(libc::S_IFDIR | 0o444))
            .expect("set_permissions failed");

        let result = create_dir(dir_path.join("sub"));

        let err = result.err().expect("create_dir returned `Ok`");
        let os_err = err.raw_os_error().expect("raw_os_error was empty");
        assert_eq!(os_err, libc::EACCES);
    }

    #[test]
    fn read_only_dir_cant_remove_subdir() {
        const DIR_NAME: &str = "etc";
        let case = TestCase::new();
        let dir_path = case.mnt_dir().join(DIR_NAME);
        let sub_path = dir_path.join("sub");
        create_dir(&dir_path).expect("create_dir failed");
        create_dir(&sub_path).expect("create_dir failed");
        set_permissions(&dir_path, Permissions::from_mode(libc::S_IFDIR | 0o444))
            .expect("set_permissions failed");

        let result = remove_dir(&sub_path);

        let err = result.err().expect("remove_dir returned `Ok`");
        let os_err = err.raw_os_error().expect("raw_os_error was empty");
        assert_eq!(os_err, libc::EACCES);
    }

    #[test]
    fn rename_file() {
        const FILE_NAME: &str = "parabola.txt";
        const EXPECTED: &[u8] = b"We are eternal all this pain is an illusion";
        let case = TestCase::new();
        let src_path = case.mnt_dir().join(FILE_NAME);
        let dst_path = case.mnt_dir().join("parabola_lyrics.txt");

        write(&src_path, EXPECTED).unwrap();
        rename(&src_path, &dst_path).unwrap();

        let actual = read(&dst_path).unwrap();
        assert_eq!(EXPECTED, actual)
    }
}
